"""
DasLaden ETL
A Simple, Tiny and Ridiculus ETL Engine made with Python

"""

__title__ = 'dasladen'
__author__ = 'Vagner Pagotti'
__author_email__ = 'pagotti@gmail.com'
__version__ = '0.1.1'
__license__ = 'MIT'
__copyright__ = 'Copyright (c) Vagner Pagotti'

from task import BaseTask
from processor import Watcher
